from .ComplexTransform import ComplexTransform
from .FourierTransform import FourierTransform
from .PowerSpectrumTransform import PowerSpectrumTransform
from .WaveletTransform import WaveletTransform
from .ZTransform import ZTransform



"""" **Description**

        Time zone interface.

    **Example**

        ::

            from diamondback import ITimeZone
            import datetime
            import pytz


            class Test( ITimeZone ) :

                def __init__( self ) -> None :

                    super( ).__init__( )

                    self.timezone = datetime.timezone.utc

            test = Test( )

            test.timezone = pytz.timezone( 'US/Eastern' )

    **License**

        `BSD-3C. <https://github.com/larryturner/diamondback/blob/master/license>`_

        Copyright (c) 2018, Larry Turner, Schneider Electric.  All rights reserved.

    **Author**

        Larry Turner, Schneider Electric, Analytics & AI, 2018-07-12.

    **Definition**

"""

from diamondback.interfaces.IEqual import IEqual
import datetime


class ITimeZone( IEqual ) :

    """ Time zone interface.
    """

    @property
    def timezone( self ) :

        """ Time zone ( timezone ).
        """

        return self._timezone

    @timezone.setter
    def timezone( self, timezone : datetime.timezone ) :

        if ( not timezone ) :

            raise ValueError( 'TimeZone = ' + str( timezone ) )

        self._timezone = timezone

    def __eq__( self, other : any ) -> bool :

        """ Equality.

            Arguments :

                other - Other ( any ).

            Returns :

                equality - Equality ( bool ).
        """

        return ( ( super( ).__eq__( other ) ) and ( self.timezone == other.timezone ) )

    def __init__( self ) -> None :

        """ Initialize.
        """

        super( ).__init__( )

        self._timezone = datetime.timezone.utc



""" **Description**

        Path interface.

    **Example**

        ::

            from diamondback import IPath

            class Test( IPath ) :

                def __init__( self ) -> None :

                    super( ).__init__( )

                    self.path = ''

            test = Test( )

            test.path = '.\data'

    **License**

        `BSD-3C. <https://github.com/larryturner/diamondback/blob/master/license>`_

        Copyright (c) 2018, Larry Turner, Schneider Electric.  All rights reserved.

    **Author**

        Larry Turner, Schneider Electric, Analytics & AI, 2020-01-09.

    **Definition**

"""

from diamondback.interfaces.IEqual import IEqual
import os

class IPath( IEqual ) :

    """ Path interface.
    """

    @property
    def path( self ) :

        """ Path ( str ).
        """

        return self._path

    @path.setter
    def path( self, path : str ) :

        if ( path ) :

            path = path.replace( '/', os.path.sep ).replace( '\\', os.path.sep )

            if ( not os.path.exists( path ) ) :

                raise FileNotFoundError( 'Path = ' + str( path ) )

        self._path = path

    def __eq__( self, other : any ) -> bool :

        """ Equality.

            Arguments :

                other - Other ( any ).

            Returns :

                equality - Equality ( bool ).
        """

        return ( ( super( ).__eq__( other ) ) and ( self.path == other.path ) )

    def __init__( self ) -> None :

        """ Initialize.
        """

        super( ).__init__( )

        self._path = [ ]



""" **Description**

        Configure interface.

    **Example**

        ::

            from diamondback import IConfigure


            class Test( IConfigure ) :

                def __init__( self ) -> None :

                    super( ).__init__( )

                    self.configure = [ ]

            test = Test( )

            test.configure = { 'a' : 0.0, 'b' : 1.0 }

    **License**

        `BSD-3C. <https://github.com/larryturner/diamondback/blob/master/license>`_

        Copyright (c) 2018, Larry Turner, Schneider Electric.  All rights reserved.

    **Author**

        Larry Turner, Schneider Electric, Analytics & AI, 2020-10-27.

    **Definition**

"""

from diamondback.interfaces.IEqual import IEqual


class IConfigure( IEqual ) :

    """ Configure interface.
    """

    @property
    def configure( self ) :

        """ Configure ( any ).
        """

        return self._configure

    @configure.setter
    def configure( self, configure : any ) :

        self._configure = configure

    def __eq__( self, other : any ) -> bool :

        """ Equality.

            Arguments :

                other - Other ( any ).

            Returns :

                equality - Equality ( bool ).
        """

        return ( ( super( ).__eq__( other ) ) and ( self.configure == other.configure ) )

    def __init__( self ) -> None :

        """ Initialize.
        """

        super( ).__init__( )

        self._configure = [ ]



from .IA import IA
from .IAsset import IAsset
from .IB import IB
from .ICache import ICache
from .IClear import IClear
from .ICompress import ICompress
from .IConfigure import IConfigure
from .IConnect import IConnect
from .ICount import ICount
from .IData import IData
from .IDate import IDate
from .IDispose import IDispose
from .IDuration import IDuration
from .IEmulate import IEmulate
from .IEncoding import IEncoding
from .IEqual import IEqual
from .IFrequency import IFrequency
from .IIdentity import IIdentity
from .IInterval import IInterval
from .ILatency import ILatency
from .ILive import ILive
from .IModel import IModel
from .IPath import IPath
from .IPeriod import IPeriod
from .IPhase import IPhase
from .IProxy import IProxy
from .IQ import IQ
from .IRate import IRate
from .IReady import IReady
from .IReset import IReset
from .IResolution import IResolution
from .IRotation import IRotation
from .IS import IS
from .IStream import IStream
from .ITimeOut import ITimeOut
from .ITimeZone import ITimeZone
from .IUpdate import IUpdate
from .IUrl import IUrl
from .IUser import IUser
from .IValid import IValid
from .IVersion import IVersion
from .IWrite import IWrite



""" **Description**

        Asset interface.

    **Example**

        ::

            from diamondback import IAsset
            import uuid


            class Test( IAsset ) :

                def __init__( self ) -> None :

                    super( ).__init__( )

                    self.asset = str( uuid.uuid4( ) )

            test = Test( )

            test.asset = '<Country-Region-City-Corporation-Site>-' + test.asset

    **License**

        `BSD-3C. <https://github.com/larryturner/diamondback/blob/master/license>`_

        Copyright (c) 2021, Larry Turner, Schneider Electric.  All rights reserved.

    **Author**

        Larry Turner, Schneider Electric, Analytics & AI, 2021-02-01.

    **Definition**

"""

from diamondback.interfaces.IEqual import IEqual


class IAsset( IEqual ) :

    """ Asset interface.
    """

    @property
    def asset( self ) :

        """ Asset ( any ).
        """

        return self._asset

    @asset.setter
    def asset( self, asset : any ) :

        self._asset = asset

    def __eq__( self, other : any ) -> bool :

        """ Equality.

            Arguments :

                other - Other ( any ).

            Returns :

                equality - Equality ( bool ).
        """

        return ( ( super( ).__eq__( other ) ) and ( self.asset == other.asset ) )

    def __init__( self ) -> None :

        """ Initialize.
        """

        super( ).__init__( )

        self._asset = [ ]



from .FirFilter import FirFilter
from .IirFilter import IirFilter
from .ComplexBandPassFilter import ComplexBandPassFilter
from .ComplexExponentialFilter import ComplexExponentialFilter
from .ComplexFrequencyFilter import ComplexFrequencyFilter
from .DerivativeFilter import DerivativeFilter
from .GoertzelFilter import GoertzelFilter
from .IntegralFilter import IntegralFilter
from .PidFilter import PidFilter
from .PolynomialRateFilter import PolynomialRateFilter
from .PolyphaseRateFilter import PolyphaseRateFilter
from .RankFilter import RankFilter
from .WindowFilter import WindowFilter




""" **Description**

        Test commons.

    **Example**

        ::

            pytest --capture=no --verbose

    **License**

        `BSD-3C. <https://github.com/larryturner/diamondback/blob/master/license>`_

        Copyright (c) 2018, Larry Turner, Schneider Electric.  All rights reserved.

    **Author**

        Larry Turner, Schneider Electric, Analytics & AI, 2018-04-03.

    **Definition**

"""

from diamondback import IirFilter
from diamondback import Log, Serial
import numpy
import os


def test_Log( count = 2, path = 'test_Log.3.14159.txt' ) :

    """ Test Log.
    """

    try :

        start, error, end = 'Log Test - Start', 'Log Test - Error', 'Log Test - End'

        with open( path, 'w' ) as fout :

            Log.stream( fout )

            Log.level( 'Error' )

            Log.write( 'Debug', error )

            Log.level( 'Debug' )

            Log.write( 'Debug', start )

            Log.write( 'Info', 'X = ', numpy.random.rand( count, count ) )

            Log.write( 'Warning', 'Y = ', list( numpy.random.rand( count, count ) ) )

            Log.write( 'Critical', 'Z = ', set( numpy.random.rand( count ) ) )

            Log.write( 'Error', ValueError( end ) )

        with open( path, 'r' ) as fin :

            x = fin.read( )

            assert x

            assert x.find( start ) >= 0

            assert x.find( error ) < 0

            assert x.find( end ) >= 0

            assert ( ( x.find( 'Critical' ) >= 0 ) and ( x.find( 'Error' ) >= 0 ) and ( x.find( 'Warning' ) >= 0 ) and ( x.find( 'Info' ) >= 0 ) and ( x.find( 'Debug' ) >= 0 ) )

    except Exception as ex :

        assert False

    finally :

        os.remove( path )


def test_Serial( count = 2 ) :

    """ Test Serial.
    """

    obj = IirFilter.Factory.instance( IirFilter, 'Butterworth', 0.1, 4, 1 )

    index = Serial.decode( Serial.encode( obj, False ), False )

    assert obj == index

    index = Serial.decode( Serial.encode( obj, True ), True )

    assert obj == index

    x = { 'array' : numpy.random.rand( count ),
          'list' : list( numpy.random.rand( count ) ),
          'set' : set( numpy.random.rand( count ) ),
          'tuple' : ( list( numpy.random.rand( count ) ) ) }

    y = Serial.decode( Serial.encode( x, False ), False )

    assert ( numpy.allclose( x[ 'array' ], y[ 'array' ] ) )

    assert ( numpy.allclose( x[ 'list' ], y[ 'list' ] ) )

    assert ( numpy.allclose( sorted( x[ 'set' ] ), sorted( y[ 'set' ] ) ) )

    assert ( numpy.allclose( x[ 'tuple' ], y[ 'tuple' ] ) )

    y = Serial.decode( Serial.encode( x, True ), True )

    assert ( numpy.allclose( x[ 'array' ], y[ 'array' ] ) )

    assert ( numpy.allclose( x[ 'list' ], y[ 'list' ] ) )

    assert ( numpy.allclose( sorted( x[ 'set' ] ), sorted( y[ 'set' ] ) ) )

    assert ( numpy.allclose( x[ 'tuple' ], y[ 'tuple' ] ) )

    x = '{ "service" : { "clear" : false, \
                         "datetime" : "2019-09-01T15:10:00Z", \
                         "duration" : 28800.0, \
                         "emulate" : true, \
                         "interval" : 3600.0, \
                         "latency" : 0.0, \
                         "level" : "info", \
                         "name" : { "one" : 1.0, "two" : 2.0, "three" : 3.14159 }, \
                         "optimalstart" : true, \
                         "optimalstop" : true, \
                         "period" : 600.0, \
                         "persistence" : 14400.0, \
                         "write" : false, \
                         "comment" : "JSON instance." } }'

    u = Serial.decode( x, False )

    assert u

    v = Serial.decode( Serial.encode( u, False ), False )

    assert len( u ) == len( v )

    key = list( u.keys( ) )[ 0 ]

    assert ( ( key in u ) and ( key in v ) )

    assert u[ key ] == v[ key ]



""" **Description**

        A diversity model realizes the selection and retention of a state as a
        finite collection of observations extracted from an incident signal, to
        maximize a minimum distance between any members of a state, according to
        a specified classification or distance metric.

        .. math::

            d_{k} = \min(\ d_{u,v}\ )\quad\quad u, v \\in [\ 0,\ M\ ),\ u \\neq v

        .. math::

            d_{k} \geq d_{n}\qquad \longrightarrow\qquad d_{n} = d_{k}

        A diversity model is an opportunistic unsupervised learning model which
        typically improves condition and numerical accuracy and reduces storage
        relative to alternative approaches including generalized linear inverse.

        A factory is defined to facilitate construction of an instance, defining
        a state array of a specified order.  A stationary dimension is inferred
        through observation.  An instance, classification, and order are specified.

        Classification is in ( 'Chebyshev', 'Euclidean', 'Geometric', 'Manhattan' ).

        * | 'Chebyshev' distance is an L-infinity norm, a maximum absolute difference
          | in any dimension.

        .. math::

            d_{u,v} = \max(\ |\ \\vec{x_{u}} - \\vec{x_{v}}\ |\ )

        * | 'Euclidean' distance is an L-2 norm, a square root of a sum of squared
          | differences in each dimension.

        .. math::

            d_{u,v} = \matrix{\sum_{i=0}^{N}(\ |\ \\vec{x_{u,i}} - \\vec{x_{v,i}}\ )^2|}^{0.5}

        * | 'Geometric' distance is a ordered root of a product of absolute differences
          | in each dimension.

        .. math::

            d_{u,v} = \prod_{i=0}^{N}{(\ |\ \\vec{x_{u,i}} - \\vec{x_{v,i}}\ |\ )}^{\\frac{1}{N}}

        * | 'Manhattan' distance is an L-1 norm, a sum of absolute differences in each
          | dimension.

        .. math::

            d_{u,v} = \sum_{i=0}^{N}{\ (\ |\ \\vec{x_{u}} - \\vec{x_{v}}\ |\ )\ }

    **Example**

        ::

            from diamondback import DiversityModel


            # Create an instance from a Factory with constraints.

            obj = DiversityModel.Factory.instance( typ = DiversityModel, classification = 'Euclidean', order = 4 )

            # Model an incident signal and extract a state.

            x = numpy.random.rand( 2, 32 )

            y = obj.model( x )

            s = obj.s

    **License**

        `BSD-3C. <https://github.com/larryturner/diamondback/blob/master/license>`_

        Copyright (c) 2018, Larry Turner, Schneider Electric.  All rights reserved.

    **Author**

        Larry Turner, Schneider Electric, Analytics & AI, 2018-02-08.

    **Definition**

"""

from diamondback.interfaces.IClear import IClear
from diamondback.interfaces.IEqual import IEqual
from diamondback.interfaces.IS import IS
import numpy
import typing


class DiversityModel( IClear, IS, IEqual ) :

    """ Diversity model.
    """

    class Factory( object ) :

        """ Factory.
        """

        _distance = { 'Chebyshev' : lambda x, y : max( abs( x - y ) ),
                      'Euclidean' : lambda x, y : sum( ( x - y ) ** 2 ) ** 0.5,
                      'Geometric' : lambda x, y : numpy.prod( abs( x - y ) ) ** ( 1.0 / len( x ) ),
                      'Manhattan' : lambda x, y : sum( abs( x - y ) ) }

        @classmethod
        def instance( cls, typ : type, classification : str, order : int ) -> any :

            """ Constructs an instance.

                Arguments :

                    typ - Type derived from DiversityModel ( type ).

                    classification - Classification in ( 'Chebyshev', 'Euclidean', 'Geometric', 'Manhattan' ) ( str ).

                    order - Order ( int ).

                Returns :

                    instance - Instance ( typ( ) ).
            """

            if ( ( not typ ) or ( not issubclass( typ, DiversityModel ) ) ) :

                raise ValueError( 'Type = ' + str( typ ) )

            if ( ( not classification ) or ( classification not in DiversityModel.Factory._distance ) ) :

                raise ValueError( 'Classification = ' + str( classification ) )

            if ( order <= 0 ) :

                raise ValueError( 'Order = ' + str( order ) )

            return typ( DiversityModel.Factory._distance[ classification ], order )

    def __eq__( self, other : any ) -> bool :

        """ Equality.

            Arguments :

                other - Other ( any ).

            Returns :

                equality - Equality ( bool ).
        """

        return ( ( super( ).__eq__( other ) ) and ( self._distance == other._distance ) and ( numpy.isclose( self._diversity, other._diversity ) ) )

    def __init__( self, distance : typing.Callable[ [ any, any ], any ], order : int ) -> None :

        """ Initialize.

            Arguments :

                distance - Distance ( method ).

                order - Order ( int ).
        """

        if ( ( not distance ) or ( isinstance( distance, str ) ) ) :

            raise ValueError( 'Distance = ' + str( distance ) )

        super( ).__init__( )

        self._distance, self._diversity, self._s = distance, 0.0, numpy.zeros( ( 0, order + 1 ) )

    def clear( self ) -> None :

        """ Clears an instance.
        """

        self._diversity, self._s = 0.0, numpy.zeros( ( 0, self.s.shape[ 1 ] ) )

    def model( self, x : any ) -> any :

        """ Models an incident signal and produces a reference signal.

            Arguments :

                x - Incident signal ( array( float ), list( float ) ).

            Returns :

                y = Diversity ( array( float ) ).
        """

        if ( ( not numpy.isscalar( x ) ) and ( not isinstance( x, numpy.ndarray ) ) ) :

            x = numpy.array( list( x ) )

        if ( ( len( x.shape ) > 2 ) or ( len( x ) == 0 ) ) :

            raise ValueError( 'X = ' + str( x ) )

        if ( len( x.shape ) < 2 ) :

            rows, cols = 1, x.shape[ 0 ]

        else :

            rows, cols = x.shape

        if ( not self.s.shape[ 0 ] ) :

            self._s = numpy.zeros( ( rows, self.s.shape[ 1 ] ) ) + numpy.finfo( float ).max

        if ( ( rows != self.s.shape[ 0 ] ) or ( cols <= 0 ) ) :

            raise ValueError( '{:30s}{:30s}'.format( 'Rows = ' + str( rows ), 'Columns = ' + str( cols ) ) )

        cc = 0

        for jj in range( 0, self.s.shape[ 1 ] ) :

            if ( numpy.isclose( self.s[ 0, jj ], numpy.finfo( float ).max ) ) :

                break

            cc += 1

        y = numpy.zeros( cols )

        for jj in range( 0, cols ) :

            if ( cc < self.s.shape[ 1 ] ) :

                self.s[ :, cc ] = x[ :, jj ]

                cc += 1

            else :

                v, ii = self._diversity, -1

                for kk in range( 0, cc ) :

                    u, s = float( 'inf' ), numpy.array( self.s )

                    s[ :, kk ] = x[ :, jj ]

                    for uu in range( 0, cc - 1 ) :

                        for vv in range( uu + 1, cc ) :

                            d = self._distance( s[ :, uu ], s[ :, vv ] )

                            if ( d < u ) :

                                u = d

                    if ( u > v ) :

                        v, ii = u, kk

                if ( v > self._diversity ) :

                    self._diversity, self.s[ :, ii ] = v, x[ :, jj ]

            y[ jj ] = self._diversity

        return y



""" **Description**

        A window filter realizes a discrete difference equation as a function
        of a forward coefficient array of a specified order, consuming an
        incident signal and producing a reference signal.

        .. math::

            y_{n} = b_{n}\ x_{n}

        A factory is defined to facilitate construction of an instance,
        defining a forward coefficient array of a specified order.  An
        instance, classification, order, and normalization are specified.

        Classification is in ( 'Blackman', 'Hamming', 'Hann', 'Kaiser' ).

        * | 'Blackman' filters demonstrate low resolution and spectral leakage
          | with improved rate of attenuation.

        * | 'Hamming' filters demonstrate minimal nearest side lobe magnitude
          | response.

        * | 'Hann' filters demonstrate high resolution and spectral leakage.

        * | 'Kaiser' filters demonstrate flexible resolution and spectral
          | leakage dependent upon a beta value of a Bessel function of the
          | first kind, with beta equal to 7.0.

        Normal condition scales a forward coefficient array to electively
        compensate for energy loss.

        .. math::

            b_{n} = b_{n}\ \\frac{ N }{ \sum_{0}^{N-1}\ |\ b_{n}\ |}

    **Example**

        ::

            from diamondback import ComplexExponentialFilter, WindowFilter
            import numpy


            # Create an instance from a Factory with constraints.

            obj = WindowFilter.Factory.instance( typ = WindowFilter, classification = 'Hann', order = 15, normal = True )

            # Filter an incident signal.

            x = ComplexExponentialFilter( 0.0 ).filter( numpy.ones( len( obj.b ) ) * 0.1 ).real

            y = obj.filter( x )

    **License**

        `BSD-3C. <https://github.com/larryturner/diamondback/blob/master/license>`_

        Copyright (c) 2018, Larry Turner, Schneider Electric.  All rights reserved.

    **Author**

        Larry Turner, Schneider Electric, Analytics & AI, 2018-04-13.

    **Definition**

"""

from diamondback.interfaces.IB import IB
from diamondback.interfaces.IEqual import IEqual
import numpy
import scipy.signal


class WindowFilter( IB, IEqual ) :

    """ Window filter.
    """

    class Factory( object ) :

        """ Factory.
        """

        _classification = ( 'Blackman', 'Hamming', 'Hann', 'Kaiser' )

        @classmethod
        def instance( cls, typ : type, classification : str, order : int, normal : bool = True ) -> any :

            """ Constructs an instance.

                Arguments :

                    typ - Type derived from WindowFilter ( type ).

                    classification - Classification in ( 'Blackman', 'Hamming', 'Hann', 'Kaiser' ) ( str ).

                    order - Order ( int ).

                    normal - Normal condition ( bool ).

                Returns :

                    instance - Instance ( typ( ) ).
            """

            if ( ( not typ ) or ( not issubclass( typ, WindowFilter ) ) ) :

                raise ValueError( 'Type = ' + str( typ ) )

            if ( ( not classification ) or ( classification not in WindowFilter.Factory._classification ) ) :

                raise ValueError( 'Classification = ' + str( classification ) )

            if ( order < 0 ) :

                raise ValueError( 'Order = ' + str( order ) )

            if ( classification == 'Kaiser' ) :

                window = ( classification.lower( ), 7.0 )

            else :

                window = classification.lower( )

            b = scipy.signal.get_window( window, order + 1, False )

            if ( normal ) :

                b *= ( order + 1 ) / sum( abs( b ) )

            return typ( b )

    def __init__( self, b : any = numpy.ones( 1 ) ) -> None :

        """ Initialize.

            Arguments :

                b - Forward coefficient ( array( float ) ).
        """

        if ( ( numpy.isscalar( b ) ) or ( len( b.shape ) != 1 ) or ( len( b ) == 0 ) or ( not b.any( ) ) ) :

            raise ValueError( 'B = ' + str( b ) )

        super( ).__init__( )

        self.b = numpy.array( b )

    def filter( self, x : any ) -> any :

        """ Filters an incident signal and produces a reference signal.

            Arguments :

                x - Incident signal ( array( complex | float ), list( complex | float ) ).

            Returns :

                y - Reference signal ( array( complex | float ) ).
        """

        if ( ( not numpy.isscalar( x ) ) and ( not isinstance( x, numpy.ndarray ) ) ) :

            x = numpy.array( list( x ) )

        if ( ( len( x.shape ) != 1 ) or ( len( x ) != len( self.b ) ) ) :

            raise ValueError( 'X = ' + str( x ) )

        return self.b * x



""" **Description**

        A complex transform converts a three-phase real signal to a complex
        signal, or converts a complex signal to a three-phase real signal, in
        equivalent and reversible representations.  A neutral condition is
        specified.

        Singleton.

        .. math::

            \gamma = \scriptsize{ \matrix{ 1 & \\frac{-1}{2} & \\frac{-1}{2}\cr 0 & \\frac{3^{0.5}}{2} & \\frac{-3^{0.5}}{2}\cr 1 & 1 & 1 } }

        .. math::

            \\alpha = \matrix{\\frac{1}{3}}^{0.5}\ e^{\ -j\ \\frac{\pi}{6}\ }

        Complex - Three-Phase.

        .. math::

            y_{n} = \matrix{ x^{T}_{n}\ \gamma\ }\ \scriptsize{[\ \matrix{ 1 & j & 0 }\ ]^{T}}

        .. math::

            \overline{\scriptsize{Neutral}}\ \qquad\longrightarrow\qquad y_{n} = y_{n}\ \\alpha

        Three-Phase - Complex.

        .. math::

            \overline{\scriptsize{Neutral}}\ \qquad\longrightarrow\qquad x_{n} = \\frac{x_{n}}{\\alpha}

        .. math::

            y_{n} = \matrix{\ \matrix{ \gamma^{T}\gamma }^{-1}\ \gamma^{T}\ \scriptsize{[\ \matrix{ real(\ x_{n}\ ) & imag(\ x_{n}\ ) & 0 }\ ]^{T}}}^{T}

    **Example** ::

        from diamondback import ComplexExponentialFilter, ComplexTransform
        import numpy


        x = ComplexExponentialFilter( 0.0 ).filter( numpy.linspace( -1.0e-4, 1.0e-4, 128 ) + 0.1 )

        # Transform an incident signal, forward and inverse.

        y = ComplexTransform.transform( x, neutral = True )

        z = ComplexTransform.transform( y, neutral = True )

    **License**

        `BSD-3C. <https://github.com/larryturner/diamondback/blob/master/license>`_

        Copyright (c) 2018, Larry Turner, Schneider Electric.  All rights reserved.

    **Author**

        Larry Turner, Schneider Electric, Analytics & AI, 2018-01-26.

    **Definition**

"""

import math
import numpy


class ComplexTransform( object ) :

    """ Complex transform.
    """

    _coefficient = ( 2.0 / 3.0 ) * numpy.array( [ [ 1.0, -0.5, -0.5 ],
                                                  [ 0.0, 0.5 * ( 3.0 ** 0.5 ), -0.5 * ( 3.0 ** 0.5 ) ],
                                                  [ 1.0, 1.0, 1.0 ] ] )

    _gain = ( ( 1.0 / 3.0 ) ** 0.5 ) * numpy.exp( -1j * math.pi / 6.0 )

    @classmethod
    def transform( cls, x : any, neutral : bool = True ) -> any :

        """ Transforms a real three-phase or complex incident signal into a complex
            or three-phase reference signal.

            Arguments :

                x - Incident signal ( array( complex | float ), list( complex | float ) ).

                neutral - Neutral or line reference condition ( bool ).

            Returns :

                y - Reference signal ( array( complex | float ) ).
        """

        if ( ( not numpy.isscalar( x ) ) and ( not isinstance( x, numpy.ndarray ) ) ) :

            x = numpy.array( list( x ) )

        if ( ( len( x.shape ) > 2 ) or ( len( x ) == 0 ) ) :

            raise ValueError( 'X = ' + str( x ) )

        if ( len( x.shape ) < 2 ) :

            rows, cols = 1, x.shape[ 0 ]

        else :

            rows, cols = x.shape

        if ( ( ( rows != 1 ) and ( rows != 3 ) ) or ( cols <= 0 ) ) :

            raise ValueError( '{:30s}{:30s}'.format( 'Rows = ' + str( rows ), 'Columns = ' + str( cols ) ) )

        if ( rows == 1 ) :

            v = x

            if ( not neutral ) :

                v = x / ComplexTransform._gain

            y = numpy.matmul( numpy.linalg.inv( ComplexTransform._coefficient ), numpy.array( [ v.real, v.imag, numpy.zeros( cols ) ] ) )

        else :

            v = numpy.matmul( ComplexTransform._coefficient, x )

            y = v[ 0, : ] + 1j * v[ 1, : ]

            if ( not neutral ) :

                y *= ComplexTransform._gain

        return y



""" **Description**

        A Proportional Integral Derivative ( PID ) filter realizes a discrete
        difference equation as a function of a forward coefficient array and a
        state array of a static order.  A forward coefficient array applies a
        gain to proportional, integral, and derivative representations of an
        incident signal, producing a reference signal.  An integral limit is
        specified, preventing integral saturation which may adversely affect
        control stability and latency.

        .. math::

            y_{n} = b_{0}\ x_{n} + b_{1}\max(\ min( \sum_{0}^{n}\ x_{n},\ limit\ ),\ -limit\ ) + b_{2}\ \\frac{d}{dn}(\ x_{n}\ )

    **Example**

        ::

            from diamondback import ComplexExponentialFilter, PidFilter
            import numpy


            # Create an instance with coefficients.

            obj = PidFilter( b = numpy.array( [ 0.1, 5.0e-2, 0.0 ] ) )

            # Filter an incident signal.

            x = ComplexExponentialFilter( 0.0 ).filter( numpy.linspace( -1.0e-4, 1.0e-4, 128 ) * 0.1 ).real

            y = obj.filter( x )

    **License**

        `BSD-3C. <https://github.com/larryturner/diamondback/blob/master/license>`_

        Copyright (c) 2018, Larry Turner, Schneider Electric.  All rights reserved.

    **Author**

        Larry Turner, Schneider Electric, Analytics & AI, 2018-01-31.

    **Definition**

"""

from diamondback.filters.FirFilter import FirFilter
import numpy


class PidFilter( FirFilter ) :

    """ Proportional Integral Derivative ( PID ) filter.
    """

    @property
    def limit( self ) :

        """ Integral limit ( float ).
        """

        return self._limit

    @limit.setter
    def limit( self, limit : float ) :

        if ( limit < 0.0 ) :

            raise ValueError( 'Limit = ' + str( limit ) )

        self._limit = limit

    def __eq__( self, other : any ) -> bool :

        """ Equality.

            Arguments :

                other - Other ( any ).

            Returns :

                equality - Equality ( bool ).
        """

        return ( ( super( ).__eq__( other ) ) and ( numpy.isclose( self.limit, other.limit ) ) )

    def __init__( self, b : any ) -> None :

        """ Initialize.

            Arguments :

                b - Forward coefficient ( array( complex | float ), list( complex | float ) ).
        """

        if ( ( not numpy.isscalar( b ) ) and ( not isinstance( b, numpy.ndarray ) ) ) :

            b = numpy.array( list( b ) )

        if ( ( len( b.shape ) != 1 ) or ( len( b ) != 3 ) ) :

            raise ValueError( 'B = ' + str( b ) )

        super( ).__init__( b, numpy.zeros( len( b ) ), 0.0 )

        self._limit = float( 'inf' )

    def filter( self, x : any, d : any = None ) -> any :

        """ Filters an incident signal and produces a reference signal.

            Arguments :

                x - Incident signal ( array( complex | float ), list( complex | float ) ).

            Returns :

                y - Reference signal ( array( complex | float ) ).
        """

        if ( ( not numpy.isscalar( x ) ) and ( not isinstance( x, numpy.ndarray ) ) ) :

            x = numpy.array( list( x ) )

        if ( ( len( x.shape ) != 1 ) or ( len( x ) == 0 ) ) :

            raise ValueError( 'X = ' + str( x ) )

        y = numpy.zeros( len( x ), type( self.b[ 0 ] ) )

        for ii in range( 0, len( x ) ) :

            self.s[ 2 ] = x[ ii ] - self.s[ 0 ]

            if ( abs( self.s[ 1 ] + x[ ii ] ) < self.limit ) :

                self.s[ 1 ] += x[ ii ]

            self.s[ 0 ] = x[ ii ]

            y[ ii ] = self.b.dot( self.s )

        return y



""" **Description**

        A complex band pass filter produces a complex exponential incident
        signal at a specified normalized frequency and adapts a forward complex
        coefficient to produce a reference signal, which estimates a component
        of interest in a primary signal.  A normalized frequency and rate of
        adaptation are specified.

        .. math::

            x_{n} = e^{\ j\ \pi\ \phi_{n}}

        .. math::

            \phi_{n+1} = \phi_{n} + f_{n}

        .. math::

            y_{n} = b_{n} x_{n}

        .. math::

            e_{n} = d_{n} - y_{n}

        .. math::

            b_{n+1} = b_{n} + \mu e_{n} x_{n}^{*}

    **Example**

        ::

            from diamondback import ComplexBandPassFilter, ComplexExponentialFilter
            import numpy


            frequency = 0.1

            x = numpy.linspace( -1.0e-4, 1.0e-4, 128 ) + frequency

            # Create a primary signal.

            d = ComplexExponentialFilter( phase = numpy.random.rand( 1 )[ 0 ] * 2.0 - 1.0 ).filter( x )

            obj = ComplexBandPassFilter( frequency = frequency, rate = 5.0e-2 )

            # Filter a primary signal.

            obj.reset( d[ 0 ] )

            y, e, b = obj.filter( d )

    **License**

        `BSD-3C. <https://github.com/larryturner/diamondback/blob/master/license>`_

        Copyright (c) 2018, Larry Turner, Schneider Electric.  All rights reserved.

    **Author**

        Larry Turner, Schneider Electric, Analytics & AI, 2018-01-31.

    **Definition**

"""

from diamondback.filters.ComplexExponentialFilter import ComplexExponentialFilter
from diamondback.filters.FirFilter import FirFilter
from diamondback.interfaces.IFrequency import IFrequency
import numpy
import typing


class ComplexBandPassFilter( FirFilter, IFrequency ) :

    """ Complex band pass filter.
    """

    def __eq__( self, other : any ) -> bool :

        """ Equality.

            Arguments :

                other - Other ( any ).

            Returns :

                equality - Equality ( bool ).
        """

        return ( ( super( ).__eq__( other ) ) and ( self._complexexponentialfilter == other._complexexponentialfilter ) )

    def __init__( self, frequency : float, rate : float ) -> None :

        """ Initialize.

            Arguments :

                frequency - Normalized frequency relative to Nyquist in [ -1.0, 1.0 ) ( float ).

                rate - Rate of adaptation in [ 0.0, 1.0 ] ( float ).
        """

        super( ).__init__( numpy.array( [ numpy.finfo( float ).eps + 0j ] ), numpy.zeros( 1, complex ), rate )

        self._complexexponentialfilter = ComplexExponentialFilter( )

        self.frequency = frequency

    def filter( self, d : any, x : any = None ) -> typing.Tuple[ any, any, any ] :

        """ Filters an incident signal and produces a reference signal.

            Arguments :

                d - Primary signal ( array( complex | float ), list( complex | float ) ).

            Returns :

                y - Reference signal ( array( complex | float ) ).

                e - Error signal ( array( complex | float ) ).

                b - Forward coefficient ( array( complex ) ).
        """

        if ( ( not numpy.isscalar( d ) ) and ( not isinstance( d, numpy.ndarray ) ) ) :

            d = numpy.array( list( d ) )

        if ( ( len( d.shape ) != 1 ) or ( len( d ) == 0 ) ) :

            raise ValueError( 'D = ' + str( d ) )

        x = self._complexexponentialfilter.filter( numpy.ones( len( d ) ) * self.frequency )

        y, e, b = numpy.zeros( len( x ), complex ), numpy.zeros( len( x ), complex ), numpy.zeros( len( x ), complex )

        for ii in range( 0, len( x ) ) :

            y[ ii ] = x[ ii ] * self.b[ 0 ]

            e[ ii ] = d[ ii ] - y[ ii ]

            b[ ii ] = self.b[ 0 ]

            self.b[ 0 ] += self.rate * e[ ii ] * numpy.conjugate( x[ ii ] )

        if ( not isinstance( d[ 0 ], complex ) ) :

            y, e = y.real, e.real

        return y, e, b



""" **Description**

        Clear interface.

    **Example**

        ::

            from diamondback import IClear, IPhase


            class Test( IClear, IPhase ) :

                def clear( self ) -> None :

                    self.phase = 0.0

            test = Test( )

            test.clear( )

    **License**

        `BSD-3C. <https://github.com/larryturner/diamondback/blob/master/license>`_

        Copyright (c) 2019, Larry Turner, Schneider Electric.  All rights reserved.

    **Author**

        Larry Turner, Schneider Electric, Analytics & AI, 2019-01-25.

    **Definition**

"""

from abc import ABC, abstractmethod


class IClear( ABC ) :

    """ Clear interface.
    """

    def __init__( self ) -> None :

        """ Initialize.
        """

        super( ).__init__( )

    @abstractmethod
    def clear( self ) -> any :

        """ Clears an instance.
        """

        pass




""" **Description**

        A serial instance encodes and decodes an instance or collection with
        JSON, or base-64 encoded gzip JSON binary format.

        An instance may be an object or a collection, referenced by abstract or
        concrete types, and the instance will be correctly encoded and decoded.
        JSON binary format is selected by electing to compress.  Encoding may
        be specified if an alternative to UTF-8 is required.

        Comments may be filtered from JSON by electing to clean.  Python style
        docstring and line comments are supported, though line comments must be
        terminated by a new line.

        Singleton.

        Thread safe.

    **Example**

        ::

            from diamondback import Serial
            import numpy
            import pandas


            # Encode and decode a dictionary instance in JSON.

            x = { 'a' : numpy.random.rand( count ),
                  'b' : list( numpy.random.rand( count ) ) }

            z = Serial.decode( Serial.encode( x ) )

            # Encode and decode a dictionary instance in gzip JSON.

            y = Serial.encode( x, compress = True )

            z = Serial.decode( y, compress = True )

            # Encode and decode a pandas data frame in gzip JSON.

            model = pandas.DataFrame( { 'Fruit' : [ 'Orange', 'Apple', 'Kiwi' ], 'Cost' : [ 1.25, 1.5, 0.30 ] } )

            z = Serial.decode( Serial.encode( x ) )

            # Decode a dictionary instance from JSON.

            z = Serial.decode( '{ "a" : 1.0, "b" : 2.0, "c" : 3.14159 }' )

            # Decode a dictionary instance from JSON, and clean comments.

            z = Serial.decode( '{ "a" : 1.0, "b" : 2.0, "c" : 3.14159 }  # Comments.\n', clean = True )

    **License**

        `BSD-3C. <https://github.com/larryturner/diamondback/blob/master/license>`_

        Copyright (c) 2018, Larry Turner, Schneider Electric.  All rights reserved.

    **Author**

        Larry Turner, Schneider Electric, Analytics & AI, 2018-07-13.

    **Definition**

"""

import base64
import gzip
import jsonpickle
import jsonpickle.ext.numpy
import jsonpickle.ext.pandas
import re


class Serial( object ) :

    """ Serial service, with JSON or base-64 encoded gzip JSON binary format.
    """

    jsonpickle.ext.numpy.register_handlers( )

    jsonpickle.ext.pandas.register_handlers( )

    @staticmethod
    def decode( state : str, compress : bool = False, encoding : str = 'utf_8', clean : bool = False ) -> any :

        """ Decodes an instance or collection from JSON, or base-64 encoded
            gzip JSON binary format state.  Encoding may be specified if an
            alternative to UTF-8 is required.  Python style docstring and line
            comments may be cleaned, though line comments must be terminated by
            a new line.

            Arguments :

                state - State ( str ).

                compress - Compress ( bool ).

                encoding - Encoding ( str ).

                clean - Clean comments ( bool ).

            Returns :

                instance - Instance ( any ).
        """

        if ( not state ) :

            raise ValueError( 'State = ' + str( state ) )

        if ( not encoding ) :

            raise ValueError( 'Encoding = ' + str( encoding ) )

        if ( compress ) :

            state = str( gzip.decompress( base64.b64decode( bytes( state, encoding ) ) ), encoding )

        if ( clean ) :

            state = re.sub( re.compile( '#.*?\n', re.DOTALL ), '', re.sub( re.compile( '\""".*?\"""', re.DOTALL ), '', state ) )

        return jsonpickle.decode( state )

    @staticmethod
    def encode( instance : any, compress : bool = False, encoding : str = 'utf_8' ) -> str :

        """ Encodes JSON, or base-64 encoded gzip JSON binary format state from
            an instance or collection.  Encoding may be specified if an
            alternative to UTF-8 is required.

            Arguments :

                instance - Instance ( any ).

                compress - Compress ( bool ).

                encoding - Encoding ( str ).

            Returns :

                state - State ( str ).
        """

        if ( not encoding ) :

            raise ValueError( 'Encoding = ' + str( encoding ) )

        state = jsonpickle.encode( instance )

        if ( compress ) :

            state = str( base64.b64encode( gzip.compress( bytes( state, encoding ) ) ), encoding )

        return state



""" **Description**

        Model interface.

    **Example**

        ::

            from diamondback import IModel, Serial

            class Test( IModel ) :

                def __init__( self ) -> None :

                    super( ).__init__( )

                    self.model = [ ]

            test = Test( )

            test.model = Serial.encode( { 'a' : 0.0, 'b' : 1.0 } )

    **License**

        `BSD-3C. <https://github.com/larryturner/diamondback/blob/master/license>`_

        Copyright (c) 2018, Larry Turner, Schneider Electric.  All rights reserved.

    **Author**

        Larry Turner, Schneider Electric, Analytics & AI, 2018-07-12.

    **Definition**

"""

from diamondback.interfaces.IEqual import IEqual


class IModel( IEqual ) :

    """ Model interface.
    """

    @property
    def model( self ) :

        """ Model ( any ).
        """

        return self._model

    @model.setter
    def model( self, model : any ) :

        self._model = model

    def __eq__( self, other : any ) -> bool :

        """ Equality.

            Arguments :

                other - Other ( any ).

            Returns :

                equality - Equality ( bool ).
        """

        return ( ( super( ).__eq__( other ) ) and ( self.model == other.model ) )

    def __init__( self ) -> None :

        """ Initialize.
        """

        super( ).__init__( )

        self._model = [ ]



""" **Description**

        Rotation interface.

    **Example**

        ::

            from diamondback import IRotation


            class Test( IRotation ) :

                def __init__( self ) -> None :

                    super( ).__init__( )

                    self.rotation = 0.0

            test = Test( )

            test.rotation = 90.0

    **License**

        `BSD-3C. <https://github.com/larryturner/diamondback/blob/master/license>`_

        Copyright (c) 2019, Larry Turner, Schneider Electric.  All rights reserved.

    **Author**

        Larry Turner, Schneider Electric, Analytics & AI, 2019-10-09.

    **Definition**

"""

from diamondback.interfaces.IEqual import IEqual
import numpy


class IRotation( IEqual ) :

    """ Rotation interface.
    """

    @property
    def rotation( self ) :

        """ Rotation ( float ).
        """

        return self._rotation

    @rotation.setter
    def rotation( self, rotation : float ) :

        self._rotation = rotation

    def __eq__( self, other : any ) -> bool :

        """ Equality.

            Arguments :

                other - Other ( any ).

            Returns :

                equality - Equality ( bool ).
        """

        return ( ( super( ).__eq__( other ) ) and ( numpy.isclose( self.rotation, other.rotation ) ) )

    def __init__( self ) -> None :

        """ Initialize.
        """

        super( ).__init__( )

        self._rotation = 0.0



""" **Description**

        User interface.

    **Example**

        ::

            from diamondback import IUser


            class Test( IUser ) :

                def __init__( self ) -> None :

                    super( ).__init__( )

            test = Test( )

            print( test.user )

    **License**

        `BSD-3C. <https://github.com/larryturner/diamondback/blob/master/license>`_

        Copyright (c) 2020, Larry Turner, Schneider Electric.  All rights reserved.

    **Author**

        Larry Turner, Schneider Electric, Analytics & AI, 2020-11-02.

    **Definition**

"""

from diamondback.interfaces.IEqual import IEqual
import getpass


class IUser( IEqual ) :

    """ User interface.
    """

    @property
    def user( self ) :

        """ User ( str ).
        """

        try :

            value = getpass.getuser( )

        except :

            value = ''

        return value

    def __eq__( self, other : any ) -> bool :

        """ Equality.

            Arguments :

                other - Other ( any ).

            Returns :

                equality - Equality ( bool ).
        """

        return ( ( super( ).__eq__( other ) ) and ( self.user == other.user ) )

    def __init__( self ) -> None :

        """ Initialize.
        """

        super( ).__init__( )



""" **Description**

        Duration interface.

    **Example**

        ::

            from diamondback import IDuration


            class Test( IDuration ) :

                def __init__( self ) -> None :

                    super( ).__init__( )

                    self.duration = 0.0

            test = Test( )

            test.duration = 3600.0

    **License**

        `BSD-3C. <https://github.com/larryturner/diamondback/blob/master/license>`_

        Copyright (c) 2018, Larry Turner, Schneider Electric.  All rights reserved.

    **Author**

        Larry Turner, Schneider Electric, Analytics & AI, 2018-07-12.

    **Definition**

"""

from diamondback.interfaces.IEqual import IEqual
import numpy


class IDuration( IEqual ) :

    """ Duration interface.
    """

    @property
    def duration( self ) :

        """ Duration in seconds ( float ).
        """

        return self._duration

    @duration.setter
    def duration( self, duration : float ) :

        if ( duration < 0.0 ) :

            raise ValueError( 'Duration = ' + str( duration ) )

        self._duration = duration

    def __eq__( self, other : any ) -> bool :

        """ Equality.

            Arguments :

                other - Other ( any ).

            Returns :

                equality - Equality ( bool ).
        """

        return ( ( super( ).__eq__( other ) ) and ( numpy.isclose( self.duration, other.duration ) ) )

    def __init__( self ) -> None :

        """ Initialize.
        """

        super( ).__init__( )

        self._duration = 0.0



""" **Description**

        A Fourier transform converts a real or complex discrete-time incident
        signal to a complex discrete-frequency reference signal, or a complex
        discrete-frequency incident signal to a real or complex discrete-time
        reference signal, in equivalent and reversible representations.  A
        forward coefficient array is specified to define a window filter.

        Singleton.

        .. math::

            y_{k} = \\frac{1}{N}\ \sum_{n = 0}^{N-1} b_{n} x_{n} e^{\\frac{\ -j\ \pi\ k \ n}{N}} = \\frac{1}{N}\ fft(\ x_{n}\ )

        .. math::

            x_{n} = \\frac{N}{b_{n}}\ \sum_{k = 0}^{N-1} y_{k} e^{\\frac{\ j\ \pi\ k \ n}{N}} = \\frac{N}{b_{n}}\ ifft(\ y_{k}\ )

        A Fourier transform is normalized by incident signal length and forms
        a contiguous sequence corresponding to a linear and increasing
        normalized frequency.

        .. math::

            f_{k} = -1\ + \ 2\ \\frac{k}{N}

        An incident signal length is inversely proportional to a normalized
        frequency resolution.

            N = \\frac{2}{R}

    **Example** ::

        from diamondback import ComplexExponentialFilter, FourierTransform
        import numpy


        x = ComplexExponentialFilter( 0.0 ).filter( numpy.linspace( 0.12, 0.23, 128 ) ) * numpy.random.rand( 1 )[ 0 ]

        b = WindowFilter.Factory.instance( WindowFilter, 'Hann', len( x ) - 1 ).b

        # Transform an incident signal, forward and inverse.

        y, f = FourierTransform.transform( x, b = b, inverse = False )

        z = FourierTransform.transform( y, b = b, inverse = True )[ 0 ]

    **License**

        `BSD-3C. <https://github.com/larryturner/diamondback/blob/master/license>`_

        Copyright (c) 2018, Larry Turner, Schneider Electric.  All rights reserved.

    **Author**

        Larry Turner, Schneider Electric, Analytics & AI, 2018-04-12.

    **Definition**

"""

import numpy
import typing


class FourierTransform( object ) :

    """ Fourier transform.
    """

    @staticmethod
    def transform( x : any, b : any, inverse : bool = False ) -> typing.Tuple[ any, any ] :

        """ Transforms a real or complex discrete-time incident signal to a
            complex discrete-frequency reference signal, or performs an
            inverse transform.  Indices definition depends upon an inverse
            condition.  Forward transform indices define normalized frequency.
            Inverse transform indices define an integral sequence.

            Arguments :

                x - Incident signal ( array( complex | float ), list( complex | float ) ).

                b - Forward coefficient ( array( float ), list( float ) ).

                inverse - Inverse condition ( bool ).

            Returns :

                y - Reference signal ( array( complex | float ) ).

                f - Normalized frequency relative to Nyquist in [ -1.0, 1.0 ) ( array( float ) ).
        """

        if ( ( not numpy.isscalar( x ) ) and ( not isinstance( x, numpy.ndarray ) ) ) :

            x = numpy.array( list( x ) )

        if ( ( len( x.shape ) != 1 ) or ( len( x ) == 0 ) ) :

            raise ValueError( 'X = ' + str( x ) )

        if ( ( not numpy.isscalar( b ) ) and ( not isinstance( b, numpy.ndarray ) ) ) :

            b = numpy.array( list( b ) )

        if ( ( len( b.shape ) != 1 ) or ( len( b ) > len( x ) ) or ( numpy.isclose( b[ 1 : -1 ], 0.0 ).any( ) ) ) :

            raise ValueError( 'B = ' + str( b ) )

        u = numpy.array( x[ : len( b ) ] )

        v = numpy.array( b )

        if ( inverse ) :

            if ( ( numpy.isclose( v[ 0 ], 0.0 ) ) or ( numpy.isclose( v[ -1 ], 0.0 ) ) ) :

                v[ 0 ], v[ -1 ] = v[ 1 ], v[ -2 ]

            y = numpy.concatenate( ( u[ len( u ) // 2 : ], u[ : len( u ) // 2 ] ) )

            y, f = numpy.fft.ifft( y ) * len( y ) / v, numpy.linspace( 0, len( y ) - 1, len( y ) )

        else :

            y, f = numpy.fft.fft( v * u ) / len( u ), numpy.linspace( -1.0, 1.0 - 2.0 / len( u ), len( u ) )

            y = numpy.concatenate( ( y[ len( y ) // 2 : ], y[ : len( y ) // 2 ] ) )

        return y, f



""" **Description**

        Update interface.

    **Example**

        ::

            from diamondback import IPhase, IUpdate


            class Test( IPhase, IUpdate ) :

                def update( self ) -> None :

                    self.phase += 1.0

            test = Test( )

            test.update( )

    **License**

        `BSD-3C. <https://github.com/larryturner/diamondback/blob/master/license>`_

        Copyright (c) 2018, Larry Turner, Schneider Electric.  All rights reserved.

    **Author**

        Larry Turner, Schneider Electric, Analytics & AI, 2018-03-23.

    **Definition**

"""

from abc import ABC, abstractmethod


class IUpdate( ABC ) :

    """ Update interface.
    """

    def __init__( self ) -> None :

        """ Initialize.
        """

        super( ).__init__( )

    @abstractmethod
    def update( self ) -> any :

        """ Update.
        """

        pass




""" **Description**

        A log instance formats and writes log entries, electively using the
        logger package or directly to a specified stream.  Log entries are
        prefaced with an ISO-8601 datetime and log level, and enhancements are
        made to the formatting of datetime, exception, and collection data
        types.  Dynamic stream redirection and log level specification are
        supported.

        If a stream is defined and associated with a specified name, the
        standard logging package is used in subsequent interactions with a log,
        a logger instance is defined with the specified name, a logging stream
        handler is associated with the stream, and log levels are synchronized
        with logging levels.

        Singleton.

        Thread safe.

    **Example**

        ::

            from diamondback import Log
            import io
            import numpy
            import pytz
            import sys


            try :

                # Default - Log stream is sys.stdout, logging package is not used, level is 'Info', and timezone is UTC.

                # Set Log level to 'Info'.

                Log.level( 'Info' )

                # Write an 'Info' entry with UTC timezone.

                Log.write( 'Info', 'Hello' )

                # Write an 'Info' entry with 'US/Eastern' timezone.

                Log.timezone( pytz.timezone( 'US/Eastern' ) )

                Log.write( 'Info', 'World', ( 'Example', 'Data', 'Payload' ) )

                # Set Log stream to sys.stdout, use logging as 'output', and write an 'Info' entry.

                Log.stream( sys.stdout, 'output' )

                Log.write( 'Info', 'Valid = ', True )

                # Set Log stream to a memory stream, write an Info entry, and read and reset the stream.

                stream = io.StringIO( )

                Log.stream( stream )

                x = numpy.random.rand( 2, 2 )

                Log.write( 'Info', 'X = ', x )

                value = stream.getvalue( )

                _, _ = stream.truncate( 0 ), stream.seek( 0 )

                # Set Log stream to a file and write a 'Warning' entry.

                with open( 'log.000.txt', 'w' ) as fout :

                    Log.stream( fout )

                    x = numpy.random.rand( 2, 2 )

                    Log.write( 'Warning', 'X = ', x )

            except Exception as ex :

                # Write an 'Error' entry on Exception.

                Log.write( ex, 'Error' )

    **License**

        `BSD-3C. <https://github.com/larryturner/diamondback/blob/master/license>`_

        Copyright (c) 2018, Larry Turner, Schneider Electric.  All rights reserved.

    **Author**

        Larry Turner, Schneider Electric, Analytics & AI, 2018-03-22.

    **Definition**

"""

from threading import RLock
import datetime
import logging
import numpy
import os
import sys
import typing


class Log( object ) :

    """ Log service.
    """

    numpy.set_printoptions( formatter = { 'float' : '{: .6f}'.format } )

    _handler = [ ]

    _level, _log = 'Info', [ ]

    _map = { 'Critical' : logging.CRITICAL,
             'Error' : logging.ERROR,
             'Warning' : logging.WARNING,
             'Info' : logging.INFO,
             'Debug' : logging.DEBUG }

    _rlock = RLock( )

    _stream, _timezone = sys.stdout, datetime.timezone.utc

    @classmethod
    def level( cls, level : str ) -> None :

        """ Level.

            Arguments :

                level - Level in ( 'Critical', 'Error', 'Warning', 'Info', 'Debug' ) ( str ).
        """

        with ( Log._rlock ) :

            if ( not level ) :

                raise ValueError( 'Level = ' + str( level ) )

            level = level.title( )

            if ( level != Log._level ) :

                if ( level not in Log._map ) :

                    raise ValueError( 'Level = ' + str( level ) )

                if ( Log._log ) :

                    Log._log.setLevel( Log._map[ level ] )

                Log._level = level

    @classmethod
    def stream( cls, stream : any, name : str = None ) -> None :

        """ Stream.

            Arguments :

                stream - Stream ( sys.stderr, sys.stdout, open( < path >, 'w' or 'a' ) ).

                name - Logger name if not empty ( str ).
        """

        with ( Log._rlock ) :

            if ( not stream ) :

                raise ValueError( 'Stream = ' + str( stream ) )

            if ( Log._log ) :

                Log._log.removeHandler( Log._handler )

                Log._handler, Log._log = [ ], [ ]

            if ( name ) :

                Log._handler = logging.StreamHandler( stream )

                Log._log = logging.getLogger( name )

                Log._log.addHandler( Log._handler )

                Log._log.setLevel( Log._map[ Log._level ] )

            Log._stream = stream

    @classmethod
    def timezone( cls, timezone : datetime.timezone ) -> None :

        """ Timezone.

            Arguments :

                timezone - Time zone ( timezone ).
        """

        with ( Log._rlock ) :

            if ( not timezone ) :

                raise ValueError( 'TimeZone = ' + str( timezone ) )

            Log._timezone = timezone

    @classmethod
    def write( cls, level : str, entry : typing.Union[ Exception, str ], data : any = None ) -> None :

        """ Formats and writes log entries, electively using the logger package
            or directly to a specified stream.  Log entries are prefaced with
            an ISO-8601 datetime and log level, and enhancements are made to
            the formatting of datetime, exception, and collection data types.

            Arguments :

                level - Level in ( 'Critical', 'Error', 'Warning', 'Info', 'Debug' ) ( str ).

                entry - Entry ( Exception, str ).

                data - Data ( any ).
        """

        with ( Log._rlock ) :

            level = level.title( )

            if ( Log._map[ level ] >= Log._map[ Log._level ] ) :

                try :

                    s = '{:30s}{:10s}'.format( datetime.datetime.utcnow( ).replace( microsecond = 0, tzinfo = datetime.timezone.utc ).astimezone( Log._timezone ).isoformat( ), level )

                    if ( isinstance( entry, Exception ) ) :

                        s += '{:30s}{:s}'.format( type( entry ).__name__, str( entry ) )

                        info = sys.exc_info( )[ -1 ]

                        while ( info ) :

                            s += os.linesep + '{:40s}@ {:s} {:s} {:d}'.format( '', info.tb_frame.f_code.co_filename.split( os.sep )[ -1 ],
                                                                                   info.tb_frame.f_code.co_name,
                                                                                   info.tb_lineno )

                            info = info.tb_next

                    else :

                        s += str( entry )

                    try :

                        if ( data is not None ) :

                            if ( len( data ) > 0 ) :

                                if ( isinstance( data, str ) ) :

                                    s += ' ' + str( data )

                                else :

                                    s += ' ' + str( data ).replace( '(', '( ' ).replace( ')', ' )' ).replace( '[', '[ ' ).replace( ']', ' ]' ).replace( '{', '{ ' ).replace( '}', ' }' )

                    except :

                        s += ' ' + str( data )

                    if ( Log._log ) :

                        Log._log.log( Log._map[ level ], s )

                    else :

                        _ = Log._stream.write( s + os.linesep )

                except :

                    pass



""" **Description**

        Diamondback setup.

    **License**

        `BSD-3C. <https://github.com/larryturner/diamondback/blob/master/license>`_

        Copyright (c) 2018, Larry Turner, Schneider Electric.  All rights reserved.

    **Author**

        Larry Turner, Schneider Electric, Analytics & AI, 2018-01-23.

    **Definition**

"""

from setuptools import find_packages, setup
import diamondback


try :

    with open( 'readme.rst', 'r' ) as fin :

        _readme = fin.read( )

except :

    _readme = ''


setup( name = 'diamondback',
       version = diamondback.__version__,
       author = 'Larry Turner',
       author_email = 'larry.turner@se.com',
       url = 'https://github.com/larryturner/diamondback',
       description = 'Diamondback digital signal processing package.',
       long_description = _readme,
       classifiers = [ 'Operating System :: OS Independent',
                       'Programming Language :: Python :: 3',
                       'License :: OSI Approved :: BSD License' ],
       install_requires = [ 'jsonpickle >= 1.4.2',
                            'numpy >= 1.19.4',
                            'pandas >= 1.2.0',
                            'requests >= 2.25.1',
                            'scipy >= 1.5.4' ],
       keywords = [ 'DSP',
                    'FFT',
                    'FIR',
                    'GZIP',
                    'IIR',
                    'JSON',
                    'PCA',
                    'PID',
                    'PSD',
                    'REST',
                    'adaptive',
                    'complex',
                    'derivative',
                    'diversity',
                    'exponential',
                    'filter',
                    'fourier',
                    'frequency',
                    'goertzel',
                    'integral',
                    'log',
                    'model',
                    'polynomial',
                    'polyphase',
                    'power',
                    'rate',
                    'serial',
                    'transform',
                    'wavelet' ],
       packages = find_packages( ),
       python_requires = '>= 3.8',
       license = 'BSD-3C.  Copyright (c) 2018, Larry Turner, Schneider Electric.  All rights reserved.' )





""" **Description**

    REST client for simple REST service requests.  An API and an elective
    dictionary of parameter strings are encoded to build a URL, elective
    JSON or binary data are defined in the body of a request, and a JSON,
    binary, or text response is returned and decoded.

    A client instance may be useful as a base client definition to interact
    with a service which satisfies flexible request constraints.

    Caching may be useful in environments with intermittent or inconsistent
    network connectivity.  If caching is specified, requests are cached when
    a service is not live, and sent in order during a subsequent request when
    a service is live.

    URL and proxy definition is supported.

    Thread safe.

    **Example**

        ::

            from diamondback import RestClient
            import requests
            import typing


            class TestClient( RestClient ) :

                def __init__( self ) -> None :

                    super( ).__init__( )

                    self.proxy = { 'http' : '', 'https' : '' }

                def add( self, item : typing.Dict[ str, float ] ) -> float :

                    return self.request( 'get', 'test/add', item = item )

            client = TestClient( )

            client.url = 'http://127.0.0.1:8080'

            client.timeout = ( 15.0, 60.0 )  # requests : ( connect, read )

            client.add( { 'x', 2.71827, 'y' : 3.14159 } )

    **License**

        `BSD-3C. <https://github.com/larryturner/diamondback/blob/master/license>`_

        Copyright (c) 2020, Larry Turner, Schneider Electric.  All rights reserved.

    **Author**

        Larry Turner, Schneider Electric, Analytics & AI, 2020-10-22.

    **Definition**

"""

from diamondback.interfaces.IClear import IClear
from diamondback.interfaces.IData import IData
from diamondback.interfaces.ILive import ILive
from diamondback.interfaces.IProxy import IProxy
from diamondback.interfaces.IReady import IReady
from diamondback.interfaces.ITimeOut import ITimeOut
from diamondback.interfaces.IUrl import IUrl
from diamondback.interfaces.IUser import IUser
from diamondback.interfaces.IVersion import IVersion
from threading import RLock
import requests
import typing


class RestClient( IClear, IData, ILive, IProxy, IReady, ITimeOut, IUrl, IUser, IVersion ) :

    """ REST client.
    """

    @ILive.live.getter
    def live( self ) :

        """ Live ( bool ).
        """

        try :

            requests.request( method = 'get', url = self.url )

            value = True

        except :

            try :

                value = requests.request( method = 'get', url = self.url + '/live' ).json( )

            except :

                value = False

        return value

    @IReady.ready.getter
    def ready( self ) :

        """ Ready ( bool ).
        """

        try :

            value = requests.request( method = 'get', url = self.url + '/ready' ).json( )

        except :

            value = False

        return value

    @IVersion.version.getter
    def version( self ) :

        """ Version ( str ).
        """

        try :

            value = requests.request( method = 'get', url = self.url + '/version' ).json( )

        except :

            value = ''

        return value

    def __init__( self ) -> None :

        """ Initialize.
        """

        super( ).__init__( )

        self._rlock = RLock( )

        self.data = [ ]

        self.proxy, self.timeout = { }, ( 15.0, 60.0 )

        self.url = 'http://127.0.0.1:8080'

    def clear( self ) -> None :

        """ Clear cached requests.
        """

        with ( self._rlock ) :

            self.data = [ ]

    def request( self, method : str, api : str, item : typing.Dict[ str, str ] = None, json : any = None, data : any = None, cache : bool = False ) -> any :

        """ Request client for simple REST service requests. An API and an
            elective dictionary of parameter strings are encoded to build a
            URL, elective JSON or binary data are defined in the body of a
            request, and a JSON, text, or binary response is returned and
            decoded.  If cache is specified, requests are cached if a service
            is not live.

            Arguments :

                method - Method ( str ) in ( 'delete', 'get', 'head', 'options', 'patch', 'post', 'put' ).

                api - API, relative to the URL ( str ).

                item - Item ( dict( str, str ) ).

                json - JSON ( any ).

                data - Data ( any ).

                cache - Cache ( bool ).

            Returns :

                value - Value ( any ).

        """

        if ( not method ) :

            raise ValueError( 'Method = ' + str( method ) )

        method = method.lower( )

        if ( method not in ( 'delete', 'get', 'head', 'options', 'patch', 'post', 'put' ) ) :

            raise ValueError( 'Method = ' + str( method ) )

        if ( ( data ) and ( json ) ) :

            raise ValueError( '{:30s}{:30s}'.format( 'Data = ' + str( data ), 'Json = ' + str( json ) ) )

        api = api.strip( '/' )

        url = self.url

        if ( api ) :

            url += '/' + api

        ready, value = True, True

        with ( self._rlock ) :

            if ( cache ) :

                if ( not self.live ) :

                    self.data.append( { 'method' : method, 'url' : url, 'item' : item, 'data' : data, 'json' : json } )

                    ready = False

            if ( ready ) :

                if ( any( self.data ) ) :

                    if ( self.live ) :

                        for x in [ x for x in self.data ] :

                            try :

                                with requests.request( method = x[ 'method' ], url = x[ 'url' ], params = x[ 'item' ], data = x[ 'data' ], json = x[ 'json' ], proxies = self.proxy, timeout = self.timeout ) as value :

                                    if ( ( not value ) or ( value.status_code != 200 ) ) :

                                        raise ConnectionError( '{:30s}{:30s}'.format( 'Status = ' + str( value.status_code ), 'Reason = ' + str( value.reason ) ) )

                            finally :

                                del self.data[ 0 ]

                with requests.request( method = method, url = url, params = item, data = data, json = json, proxies = self.proxy, timeout = self.timeout ) as value :

                    if ( ( not value ) or ( value.status_code != 200 ) ) :

                        raise ConnectionError( '{:30s}{:30s}'.format( 'Status = ' + str( value.status_code ), 'Reason = ' + str( value.reason ) ) )

                    try :

                        value = value.json( )

                    except :

                        try :

                            value = value.content

                        except :

                            value = value.text

        return value



""" **Description**

        A complex frequency filter adapts and discriminates the phase of a forward
        complex coefficient to produce a reference signal, which estimates a normalized
        frequency of a primary signal which is normalized to unity magnitude.  A
        normalized frequency and rate of adaptation are specified.

        .. math::

            f_{n} = \\frac{\\tan^{-1}(\ b_{n}\ ) }{\pi}

        .. math::

            x_{n} = \\frac{d_{n}}{|\ d_{n}\ |}

        .. math::

            y_{n} = b_{n} x_{n-1}

        .. math::

            e_{n} = d_{n} - y_{n}

        .. math::

            b_{0} = e^{\ j\ \pi\ f_{0}}

        .. math::

            b_{n} = b_{n} + \mu e_{n} x_{n}^{*}

    **Example** ::

        from diamondback import ComplexExponentialFilter
        import numpy


        x = numpy.linspace( 0.0, 0.1, 128 )

        # Create a primary signal.

        d = ComplexExponentialFilter( 0.0 ).filter( x )

        # Create an instance with frequency and rate.

        obj = ComplexFrequencyFilter( frequency = 0.0, rate = 0.1 )

        # Filter a primary signal.

        obj.reset( d[ 0 ] )

        y, e, b = obj.filter( d )

    **License**

        `BSD-3C. <https://github.com/larryturner/diamondback/blob/master/license>`_

        Copyright (c) 2018, Larry Turner, Schneider Electric.  All rights reserved.

    **Author**

        Larry Turner, Schneider Electric, Analytics & AI, 2018-02-01.

    **Definition**

"""

from diamondback.filters.FirFilter import FirFilter
from diamondback.interfaces.IFrequency import IFrequency
import math
import numpy
import typing


class ComplexFrequencyFilter( FirFilter, IFrequency ) :

    """ Complex frequency filter.
    """

    @IFrequency.frequency.setter
    def frequency( self, frequency : float) :

        """ Normalized frequency relative to Nyquist in [ -1.0, 1.0 ] ( float ).
        """

        IFrequency.frequency.fset( self, frequency )

        self.b[ 0 ] = numpy.exp( 1j * math.pi * self.frequency )

    def __init__( self, frequency : float, rate : float ) -> None :

        """ Initialize.

            Arguments :

                frequency - Normalized frequency relative to Nyquist in [ -1.0, 1.0 ) ( float ).

                rate = Rate of adaptation in [ 0.0, 1.0 ] ( float ).
        """

        super( ).__init__( numpy.ones( 1, complex ), numpy.ones( 1, complex ), rate )

        self.frequency = frequency

    def filter( self, d : any, x : any = None ) -> typing.Tuple[ any, any, any ] :

        """ Filters an incident signal and produces a reference signal.

            Arguments :

                d = Primary signal ( array( complex ), list( complex ) ).

            Returns :

                y - Reference signal ( array( float ) ).

                e = Error signal ( array( complex ) ).

                b - Forward coefficient ( array( complex ) ).
        """

        if ( ( not numpy.isscalar( d ) ) and ( not isinstance( d, numpy.ndarray ) ) ) :

            d = numpy.array( list( d ) )

        if ( ( len( d.shape ) != 1 ) or ( len( d ) == 0 ) or ( not isinstance( d[ 0 ], complex ) ) ) :

            raise ValueError( 'D = ' + str( d ) )

        x = abs( d )

        x[ numpy.isclose( x, 0.0 ) ] = 1.0

        x = d / x

        y, e, b = numpy.zeros( len( x ) ), numpy.zeros( len( x ), complex ), numpy.zeros( len( x ), complex )

        for ii in range( 0, len( x ) ) :

            y[ ii ] = numpy.angle( self.b[ 0 ] ) / math.pi

            e[ ii ] = x[ ii ] - self.b[ 0 ] * self.s[ 0 ]

            b[ ii ] = self.b[ 0 ]

            self.b[ 0 ] += self.rate * e[ ii ] * numpy.conjugate( self.s[ 0 ] )

            self.s[ 0 ] = x[ ii ]

        return y, e, b

    def reset( self, x : any ) -> None :

        """ Modifies a state to minimize edge effects by assuming persistent
            operation at a specified primary incident condition.

            Arguments :

                x - Incident signal ( complex ).
        """

        if ( not numpy.isscalar( x ) ) :

            raise ValueError( 'X = ' + str( x ) )

        if ( numpy.isclose( x, 0.0 ) ) :

            self.s[ 0 ] = 1.0

        else :

            self.s[ 0 ] = x / abs( x )
